"""Algorithms for finding motif counts in graphs"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.generic.motif."


class GlobalClusteringCoefficient(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "GlobalClusteringCoefficient"


class GlobalTriangleCount(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "GlobalTriangleCount"


class LocalClusteringCoefficient(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "LocalClusteringCoefficient"


class LocalTriangleCount(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "LocalTriangleCount"


class SquareCount(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "SquareCount"


class ThreeNodeMotifs(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "ThreeNodeMotifs"



"""Implicit schema providers used internally"""

from pyraphtory.interop import ScalaClassProxy


class SchemaProviders(ScalaClassProxy):
    _classname = "com.raphtory.internals.communication.SchemaProviderInstances"

    @classmethod
    def get_schema(cls, obj):
        return cls.generic_schema_provider()



"""Algorithms for gametheory on graphs"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.generic.gametheory."


class PrisonersDilemma(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "PrisonersDilemma"



"""Python wrappers for spouts (used for reading data with custom parse functions)"""

from pyraphtory.interop import ScalaClassProxy


"""Spouts"""


class Spout:
    """Base class for Raphtory spouts."""
    pass


class FileSpout(ScalaClassProxy, Spout):
    _classname = "com.raphtory.spouts.FileSpout"


class IdentitySpout(ScalaClassProxy, Spout):
    _classname = "com.raphtory.spouts.IdentitySpout"


class StaticGraphSpout(ScalaClassProxy, Spout):
    _classname = "com.raphtory.spouts.StaticGraphSpout"


class WebSocketSpout(ScalaClassProxy, Spout):
    _classname = "com.raphtory.spouts.WebSocketSpout"



from pyraphtory._codegen._codegen import LazyStr


def local():
    from pyraphtory.api.context import PyRaphtory
    return PyRaphtory.local()


def _local_docs():
    from pyraphtory.api.context import PyRaphtory
    return str(PyRaphtory.local.__doc__)


local.__doc__ = LazyStr(initial=lambda: _local_docs())


def remote(host: str = None, port: int = None):
    from pyraphtory.api.context import PyRaphtory
    kwargs = {}

    if host is not None:
        kwargs["host"] = host

    if port is not None:
        kwargs["port"] = port

    return PyRaphtory.remote(**kwargs)


def _remote_docs():
    from pyraphtory.api.context import PyRaphtory
    return str(PyRaphtory.remote.__doc__)


remote.__doc__ = LazyStr(initial=lambda: _remote_docs())


from pathlib import Path


build_folder = Path(__file__).resolve().parent
ivy_folder = build_folder / "ivy_data"
package_folder = build_folder.parent / "src" / "pyraphtory"
root_folder = build_folder.parent.parent
lib_folder = package_folder / "lib"
jre_folder = package_folder / "jre"
ivy_bin = package_folder / "ivy"


"""Generic algorithms that work for both multilayer and reduced views of the graph"""

from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits

_prefix = "com.raphtory.algorithms.generic."


class AdjPlus(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "AdjPlus"


class BinaryDiffusion(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "BinaryDiffusion"


class CBOD(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "CBOD"


class ConnectedComponents(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "ConnectedComponents"


class Coreness(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Coreness"


class EdgeList(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "EdgeList"


class GraphState(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "GraphState"


class HITS(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "HITS"


class KCore(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "KCore"


class LargestConnectedComponent(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "LargestConnectedComponent"


class MaxFlow(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "MaxFlow"


class NeighbourNames(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "NeighbourNames"


class NodeEdgeCount(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "NodeEdgeCount"


class NodeInformation(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "NodeInformation"


class NodeList(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "NodeList"


class TaintTrackingWithHistory(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "TaintTrackingWithHistory"


class TwoHopPaths(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "TwoHopPaths"


class VertexHistogram(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "VertexHistogram"


"""Python wrappers for Raphtory"""
import sys
from pyraphtory._context import local, remote
from pyraphtory._config import add_classpath, get_java_args, set_java_args

if sys.version_info[:2] >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

__version__ = metadata.version(__package__)
__all__ = ["local", "remote", "add_classpath", "get_java_args", "set_java_args"]



from jpype import JImplements, JOverride, JString, JObject, JClass, imports
from pyraphtory.interop._debug import enable_pydev_debug
from traceback import print_exc

Interpreter = JClass("pemja.core.Interpreter")
from java.util import Map

_globals = globals()

@JImplements(Interpreter)
class JPypeInterpreter:

    @JOverride
    def set(self, name: JString, value: JObject):
        enable_pydev_debug()
        try:
            globals()[name] = value
        except Exception as e:
            print_exc()
            raise e

    @JOverride
    def get(self, *args):
        enable_pydev_debug()
        try:
            obj = globals()[args[0]]
            if len(args) == 1:
                return obj
            else:
                return JObject(obj, args[1])
        except Exception as e:
            print_exc()
            raise e

    @JOverride
    def invoke(self, *args):
        enable_pydev_debug()
        try:
            fun = globals()[args[0]]
            if isinstance(args[1], Map):
                return fun(**args[1])
            else:
                if len(args) > 2:
                    return fun(*args[1], **args[2])
                else:
                    return fun(*args[1])
        except Exception as e:
            print_exc()
            raise e

    @JOverride
    def invokeMethod(self, obj: JString, method: JString, args):
        enable_pydev_debug()
        try:
            obj = globals()[obj]
            method = getattr(obj, str(method))
            result = method(*args)
        except Exception as e:
            print_exc()
            raise e
        return result

    @JOverride
    def exec(self, code: JString):
        enable_pydev_debug()
        try:
            exec(str(code), globals(), globals())
        except Exception as e:
            print_exc()
            raise e
        return

    @JOverride
    def close(self):
        pass



"""Algorithms for dynamics on graphs"""

from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits


_prefix = "com.raphtory.algorithms.generic.dynamic."


class DiscreteSI(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "DiscreteSI"


class Node2VecWalk(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Node2VecWalk"


class RandomWalk(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "RandomWalk"


class WattsCascade(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "WattsCascade"


class WeightedRandomWalk(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "WeightedRandomWalk"



"""Algorithms for computing shortest-paths distance"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits


_prefix = "com.raphtory.algorithms.generic.distance."


class ShortestPathDistance(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "ShortestPathDistance"



"""Python wrappers for global GraphState api"""

from pyraphtory.interop import GenericScalaProxy, register
from pyraphtory.api.scala.implicits.bounded import Bounded
from pyraphtory.api.scala.implicits.numeric import Long, Double

@register(name="Accumulator")
class Accumulator(GenericScalaProxy):
    _classname = "com.raphtory.api.analysis.graphstate.Accumulator"

    def __iadd__(self, other):
        self._plus_eq(other)
        return self


@register(name="GraphState")
class GraphState(GenericScalaProxy):
    _classname = "com.raphtory.api.analysis.graphstate.GraphState"
    """Wrapper for the global GraphState object which enables access to graphstate using getitem syntax"""
    def __getitem__(self, key):
        return self.apply(key)

    def new_int_max(self, *args, **kwargs):
        return super().new_max[Long, Bounded.long_bounds()](*args, **kwargs)

    def new_float_max(self, *args, **kwargs):
        return super().new_max[Double, Bounded.double_bounds()](*args, **kwargs)

    def new_int_min(self, *args, **kwargs):
        return super().new_min[Long, Bounded.long_bounds()](*args, **kwargs)

    def new_float_min(self, *args, **kwargs):
        return super().new_min[Double, Bounded.double_bounds()](*args, **kwargs)



"""Base classes for writing algorithms"""
from __future__ import annotations

from pyraphtory.interop import GenericScalaProxy


class ScalaAlgorithm(GenericScalaProxy):
    _classname = "com.raphtory.api.analysis.algorithm.BaseAlgorithm"


class PyAlgorithm(object):
    """
    Base class for algorithms implemented in python
    """
    def __call__(self, graph: graph.TemporalGraph) -> graph.TemporalGraph:
        """
        Main algorithm step (default implementation leaves graph unchanged)

        :param graph: input graph
        :return: transformed graph
        """
        return graph

    def tabularise(self, graph: graph.TemporalGraph) -> table.Table:
        """
        Defines the default output of the algorithm

        (default implementation includes all vertex state in the table)

        :param graph: output graph after algorithm step was applied
        :return: Table with output from this algorithm
        """
        return graph.select()

    def name(self):
        return self.__class__.__name__



"""Algorithms for community detection"""

from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits


_prefix = "com.raphtory.algorithms.generic.community."


class LPA(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "LPA"


class SLPA(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "SLPA"



"""Python wrappers for scala collections"""

from collections import abc
from pyraphtory.interop import register, GenericScalaProxy, ScalaClassProxy, to_python, to_jvm
import typing as t

A = t.TypeVar("A", covariant=True)
@register(name="Iterable")
class Iterable(GenericScalaProxy, abc.Iterable[A]):
    """Proxy object that converts scala iterables into python iterables"""
    _classname = "scala.collection.Iterable"

    def __iter__(self) -> t.Iterator[A]:
        return self.iterator()

    def __len__(self) -> int:
        return self.size()


@register(name="Iterator")
class Iterator(GenericScalaProxy, abc.Iterator[A]):
    """Proxy object that converts scala iterators into python iterators"""

    _classname = "scala.collection.Iterator"

    def __next__(self) -> A:
        if self.has_next():
            return self.next()
        else:
            raise StopIteration


@register(name="Sequence")
class Sequence(Iterable, abc.Sequence[A]):
    """Proxy object that converts scala sequence into python sequence"""
    def __len__(self) -> int:
        return self.size()

    def __getitem__(self, item: int) -> A:
        if isinstance(item, slice):
            indices = item.indices(len(self))
            if indices[2] != 1:
                # Scala does not support slice with step!
                return [self.apply(i) for i in range(*indices)]
            return self.slice(indices[0], indices[1])
        else:
            if item < 0:
                # Scala does not support negative index
                return self.apply(len(self)+item)
            else:
                return self.apply(item)

    def __contains__(self, item: A) -> bool:
        return self.contains(item)

    def __reversed__(self) -> t.Iterator[A]:
        return self.reverseIterator()

    def index(self, value: A, start: int = None, stop: int = None) -> int:
        if start < 0:
            start = len(self) + start
        if stop < 0:
            stop = len(self) + stop
        if stop is not None:
            return self.take(stop).index(value, start)
        elif start is not None:
            return self.indexOf(value, start)
        else:
            return self.indexOf(value)


K = t.TypeVar("K")
V = t.TypeVar("V", covariant=True)
@register(name="Mapping")
class Mapping(ScalaClassProxy, abc.Mapping[K, V]):
    """Proxy object that converts scala Map into python Mapping"""
    _classname = "scala.collection.Map"

    def __len__(self) -> int:
        return self.size()

    def __iter__(self) -> t.Iterator[K]:
        return self.keys_iterator()

    def __getitem__(self, key: K) -> V:
        try:
            return self.apply(key)
        except Exception:
            raise KeyError(f"missing key {key}")


class List(ScalaClassProxy, Sequence[A]):
    """Proxy object that converts scala List into python Sequence"""
    _classname = "scala.collection.immutable.List"

    @classmethod
    def _build_from_python(cls, iterable: t.Iterable[A]):
        builder = cls.new_builder()
        for element in iterable:
            builder.append(element)
        return builder.to_list()


T = t.TypeVar("T")
@register(name="Array")
class Array(GenericScalaProxy, abc.Sequence[T]):
    """Proxy object for wrapping java arrays"""
    _classname = "scala.Array"

    def __getitem__(self, item: int) -> T:
        return to_python(self.jvm[item])

    def __setitem__(self, index: int, value: T) -> None:
        self.jvm[index] = to_jvm(value)

    def __len__(self) -> int:
        return len(self.jvm)


class ScalaNone(ScalaClassProxy):
    """Convert Scala `None` to python `None`"""
    _classname = "scala.None"

    @classmethod
    def _from_jvm(cls, jvm_object: t.Any) -> None:
        return None

    def __new__(cls, jvm_object=None):
        return None


class ScalaSome(ScalaClassProxy, t.Generic[A]):
    """Proxy for unwrapping scala Some() values"""
    _classname = "scala.Some"

    @classmethod
    def _from_jvm(cls, jvm_object):
        return to_python(jvm_object.get())



"""Algorithms that explore temporal graphs"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.temporal."


class Ancestors(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Ancestors"


class Descendants(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Descendants"


class TemporalEdgeList(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "TemporalEdgeList"


class TemporalNodeList(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "TemporalNodeList"





from pyraphtory._config import jars, java, java_args, get_local_lib, jre
import os
import sys
from pyraphtory import __version__
import subprocess
import sys


def standalone():
    if java_args:
        subprocess.run([java, java_args, "-cp", jars, "com.raphtory.service.Standalone"])
    else:
        subprocess.run([java, "-cp", jars, "com.raphtory.service.Standalone"])


def clustermanager():
    if java_args:
        subprocess.run([java, java_args, "-cp", jars, "com.raphtory.service.ClusterManager"])
    else:
        subprocess.run([java, "-cp", jars, "com.raphtory.service.ClusterManager"])


def ingestion():
    if java_args:
        subprocess.run([java, java_args, "-cp", jars, "com.raphtory.service.Ingestion"])
    else:
        subprocess.run([java, "-cp", jars, "com.raphtory.service.Ingestion"])


def partition():
    if java_args:
        subprocess.run([java, java_args, "-cp", jars, "com.raphtory.service.Partition"])
    else:
        subprocess.run([java, "-cp", jars, "com.raphtory.service.Partition"])


def query():
    if java_args:
        subprocess.run([java, java_args, "-cp", jars, "com.raphtory.service.Query"])
    else:
        subprocess.run([java, "-cp", jars, "com.raphtory.service.Query"])


def classpath():
    sys.stdout.write(jars)


def version():
    sys.stdout.write(__version__)


def include():
    sys.stdout.write(str(get_local_lib()))


if __name__ == "__main__":
    standalone()



from __future__ import annotations


import subprocess
from pathlib import Path

from paths import lib_folder, ivy_folder, root_folder


ivy_files = (
    "arrow-core_2.13.jar",
    "arrow-core_2.13_ivy.xml",
    "arrow-messaging_2.13.jar",
    "arrow_messaging_2.13_ivy.xml",
    "core_2.13_ivy.xml",
    "core_2.13.jar",
)


def check_folder(folder: Path, files) -> bool:
    return all((folder / file).is_file() for file in files)


def check_build():
    build_done = check_folder(ivy_folder, ivy_files)
    return build_done


def make_python_build(rebuild=False):
    if rebuild or not check_build():
        subprocess.check_call(["make", "sbt-build"], cwd=root_folder)



from __future__ import annotations
from pathlib import Path
import hashlib
import requests
from requests.adapters import HTTPAdapter, Retry
from dataclasses import dataclass
import logging
import shutil


@dataclass
class Link:
    link: str
    checksum: str


def checksum(filepath: Path | str, expected_sha_hash: str) -> bool:
    sha256_hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    if sha256_hash.hexdigest() == expected_sha_hash:
        logging.info(f"{filepath} hash correct")
        return True
    else:
        logging.info(f"{filepath} hash invalid, got {sha256_hash.hexdigest()} expected {expected_sha_hash}")
        return False


def delete_source(filename: Path | str) -> None:
    logging.info(f"Deleting source file {filename}...")
    filename = Path(filename)
    filename.unlink(missing_ok=True)


def empty_folder(folder: str | Path):
    folder = Path(folder)
    for filename in folder.iterdir():
        try:
            if filename.is_file() or filename.is_symlink():
                filename.unlink()
            elif filename.is_dir():
                shutil.rmtree(filename, ignore_errors=True)
        except Exception:
            # silent continue
            continue


def safe_download_file(download_dir: Path | str, link: Link) -> Path:
    url = link.link
    expected_sha = link.checksum
    download_dir = Path(download_dir)
    logging.info(f"Downloading {url} to {download_dir}")
    req_session = requests.Session()
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    req_session.mount('https://', HTTPAdapter(max_retries=retries))
    filename = url.split('/')[-1]
    file_location = download_dir / filename
    # stream file to reduce mem footprint
    try:
        r = req_session.get(url, stream=True)
        # raise an exception if http issue occurs
        r.raise_for_status()
        download_dir.mkdir(parents=True, exist_ok=True)
        with open(file_location, 'wb') as f:
            shutil.copyfileobj(r.raw, f)
        status = checksum(file_location, expected_sha)
        r.close()
        if not status:
            delete_source(file_location)
            logging.info(f"Downloaded Jar {file_location} has incorrect checksum")
            raise SystemExit()
    except requests.exceptions.RequestException as err:
        logging.info(f"Major exception: {err}")
        delete_source(file_location)
        raise SystemExit(err)
    return file_location


"""Python wrappers for the table api"""

from pyraphtory.interop import GenericScalaProxy, register, logger, to_jvm
import pandas as pd
import datetime as dt


@register(name="ProgressTracker")
class ProgressTracker(GenericScalaProxy):
    _classname = "com.raphtory.api.progresstracker.ProgressTracker"

    def inner_tracker(self):
        logger.trace("Progress tracker inner tracker returned")
        return to_jvm(self)


@register(name="Table")
class Table(GenericScalaProxy):
    _classname = "com.raphtory.api.analysis.table.Table"

    def to_df(self):
        rows = []
        columns = ()
        time_formatted=False
        for res in self.get():
            columns = res.actual_header()
            window = res.perspective().window()
            timestamp = res.perspective().formatted_time()
            time_formatted = res.perspective().format_as_date()
            if window != None:
                columns = ('timestamp', 'window', *columns)
                for r in res.rows():
                    window_size = window.get().output()
                    rows.append((timestamp, window_size, *r.values()))
            else:
                columns = ('timestamp', *columns)
                for r in res.rows():
                    rows.append((timestamp, *r.values()))

        df = pd.DataFrame.from_records(rows, columns=columns)
        if time_formatted:
            df["timestamp"] = df["timestamp"].apply(lambda x: dt.datetime.strptime(x,"%Y-%m-%d %H:%M:%S.%f"))
            return df
        else:
            return df


from __future__ import annotations
from check_platform import OS_MAC, OS_LINUX, OS_AARCH64, OS_X64, getOS, getArch
from download import Link, safe_download_file, empty_folder
import logging
from pathlib import Path
import tempfile
import shutil
import os


from paths import jre_folder

SOURCES = {
    OS_MAC: {
        OS_X64: Link(
            link='https://github.com/adoptium/temurin11-binaries/releases/download/jdk-11.0.16.1%2B1/OpenJDK11U-jre_x64_mac_hotspot_11.0.16.1_1.tar.gz',
            checksum='10be61a8dd3766f7c12e2e823a6eca48cc6361d97e1b76310c752bd39770c7fe'
        ),
        OS_AARCH64: Link(
            link='https://github.com/adoptium/temurin11-binaries/releases/download/jdk-11.0.16.1%2B1/OpenJDK11U-jre_aarch64_mac_hotspot_11.0.16.1_1.tar.gz',
            checksum='c84f38a7d87d50649ffc1f625facb4398fa54885371336a2cbf6ae2b435cbd10'
        )
    },
    OS_LINUX: {
        OS_X64: Link(
            link='https://github.com/adoptium/temurin11-binaries/releases/download/jdk-11.0.17%2B8/OpenJDK11U-jre_x64_linux_hotspot_11.0.17_8.tar.gz',
            checksum='752616097e09d7f60a3ad8bd312f90eaf50ac72577e55df229fe6e8091148f79'
        ),
        OS_AARCH64: Link(
            link='https://github.com/adoptium/temurin11-binaries/releases/download/jdk-11.0.17%2B8/OpenJDK11U-jre_aarch64_linux_hotspot_11.0.17_8.tar.gz',
            checksum='bd6efe3290c8b5a42f695a55a26f3e3c9c284288574879d4b7089f31f5114177'
        )
    },
    # 'Windows': {
    #     OS_X64:
    #         {
    #             LINK: 'https://github.com/adoptium/temurin11-binaries/releases/download/jdk-11.0.17%2B8/OpenJDK11U-jre_x64_windows_hotspot_11.0.17_8.zip',
    #             CHECKSUM_SHA256: '814a731f92dd67ad6cfb11a8b06dfad5f629f67be88ae5ae37d34e6eea6be6f4'
    #         }
    # }
}


def download_java(system_os: str, architecture: str, download_dir: str | Path) -> Path:
    # get the version and sha which is stored locally
    if system_os not in SOURCES:
        msg = f'Error: System OS {system_os} not supported.\nNot downloading.'
        logging.info(msg)
        raise SystemExit(msg)
    link = SOURCES[system_os][architecture]
    # open a session and download
    logging.info(f"Downloading JAVA JRE {system_os}-{architecture} from {link.link}. Please wait...")
    file_location = safe_download_file(download_dir, link)
    return file_location


def unpack_jre(filename: str | Path, jre_loc: str | Path):
    filename = Path(filename)
    jre_loc = Path(jre_loc)
    with tempfile.TemporaryDirectory() as unpack_dir:
        unpack_dir = Path(unpack_dir)
        system_os = getOS()
        logging.info(f'Unpacking JRE for {system_os}...')
        shutil.unpack_archive(filename, unpack_dir)
        try:
            result_dir = next(unpack_dir.iterdir())
        except StopIteration:
            raise Exception('Error: JRE unpacking failed.')
        # Empty jre folder
        empty_folder(jre_loc)
        if system_os == OS_MAC:
            move_dir = 'Contents/Home'
        else:  # if system_os == OS_LINUX:
            move_dir = '.'
        for file_name in (result_dir / move_dir).iterdir():
            shutil.move(file_name, jre_loc)
        logging.info('Cleaning up...')


def check_system_dl_java() -> Path:
    java = jre_folder / "bin" / 'java'
    if not java.is_file():
        with tempfile.TemporaryDirectory() as download_dir:
            download_dir = Path(download_dir)
            logging.info("Downloading java...")
            system_os = getOS()
            logging.info(f"- Operating system: {system_os} ")
            architecture = getArch()
            logging.info(f"- Architecture: {architecture}")
            logging.info(f"Downloading to {download_dir}")
            jre_folder.mkdir(exist_ok=True, parents=True)
            download_loc = download_java(system_os, architecture, download_dir)
            unpack_jre(download_loc, jre_folder)
    return java



"""Output formats for use with sinks"""

from pyraphtory.interop import ScalaClassProxy, GenericScalaProxy


class Format(GenericScalaProxy):
    _classname = "com.raphtory.api.output.format.Format"


class JsonFormat(ScalaClassProxy, Format):
    _classname = "com.raphtory.formats.JsonFormat"


class CsvFormat(ScalaClassProxy, Format):
    _classname = "com.raphtory.formats.CsvFormat"



"""Python wrappers for sources (used to create graphs from files and databases)"""

from pyraphtory.interop import ScalaClassProxy

class Source(ScalaClassProxy):
    _classname = "com.raphtory.api.input.Source"

class CSVEdgeListSource(ScalaClassProxy):
    _classname = "com.raphtory.sources.CSVEdgeListSource"

class JSONEdgeListSource(ScalaClassProxy):
    _classname = "com.raphtory.sources.JSONEdgeListSource"

class JSONSource(ScalaClassProxy):
    _classname = "com.raphtory.sources.JSONSource"

class SqlEdgeSource(ScalaClassProxy):
    _classname = "com.raphtory.sources.SqlEdgeSource"

class SqlVertexSource(ScalaClassProxy):
    _classname = "com.raphtory.sources.SqlVertexSource"

class SqliteConnection(ScalaClassProxy):
    _classname = "com.raphtory.sources.SqliteConnection"

class PostgresConnection(ScalaClassProxy):
    _classname = "com.raphtory.sources.PostgresConnection"


"""
Entrypoints for local and remote Raphtory contexts.

Contexts are used to create and manage different graphs.
"""


from pyraphtory.interop import ScalaClassProxy, logger, ScalaPackage


class PyRaphtory(ScalaClassProxy):
    _classname = "com.raphtory.internals.context.PyRaphtoryContext"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        logger.debug("PyRaphtory context closed using context manager")

    algorithms = ScalaPackage("com.raphtory.algorithms")



"""Python wrappers for the vertex api"""

from pyraphtory.interop import GenericScalaProxy, register
from pyraphtory.api.scala.implicits.schemas import SchemaProviders


@register(name="Vertex")
class Vertex(GenericScalaProxy):
    """Wrapper for Raphtory vertex with setitem and getitem methods for working with vertex state"""

    _classname = "com.raphtory.api.analysis.visitor.Vertex"

    def message_vertex(self, vertex_id, data):
        super().message_vertex[SchemaProviders.get_schema(data)](vertex_id, data)

    def message_self(self, data):
        super().message_self[SchemaProviders.get_schema(data)](data)

    def message_out_neighbours(self, message):
        super().message_out_neighbours[SchemaProviders.get_schema(message)](message)

    def message_in_neighbours(self, message):
        super().message_in_neighbours[SchemaProviders.get_schema(message)](message)

    def message_all_neighbours(self, message):
        super().message_all_neighbours[SchemaProviders.get_schema(message)](message)
          
    def __setitem__(self, key, value):
        self.set_state(key, value)

    def __contains__(self, item):
        return self.contains_state(item, True)

    def __getitem__(self, key):
        if key in self:
            return self.get_state(key, True)
        else:
            raise KeyError(f"State with {key=} does not exist.")






"""Algorithms for community detection"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.temporal.community."


class MultilayerLPA(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "MultilayerLPA"






"""Wrappers for providing implicit numeric value required by accumulators"""

from pyraphtory.interop import ScalaClassProxy


class Long(ScalaClassProxy):
    """Wrapper for Long implicit instance"""
    _classname = "scala.math.Numeric.LongIsIntegral"


class Int(ScalaClassProxy):
    """Wrapper for Int implicit instance"""
    _classname = "scala.math.Numeric.IntIsIntegral"


class Double(ScalaClassProxy):
    """Wrapper for Double implicit instance"""
    _classname = "scala.math.Numeric.DoubleIsFractional"


class Float(ScalaClassProxy):
    """Wrapper for Float implicit instance"""
    _classname = "scala.math.Numeric.FloatIsFractional"



"""Algorithms for filtering vertices and edges"""

from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits


_prefix = "com.raphtory.algorithms.filters."

class DisparityFilter(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "DisparityFilter"


class EdgeFilter(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "EdgeFilter"


class EdgeFilterGraphState(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "EdgeFilterGraphState"


class EdgeQuantileFilter(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "EdgeQuantileFilter"


class LargestConnectedComponentFilter(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "LargestConnectedComponentFilter"


class UniformEdgeSample(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "UniformEdgeSample"


class UniformVertexSample(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "UniformVertexSample"


class VertexFilter(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "VertexFilter"


class VertexFilterGraphState(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "VertexFilterGraphState"


class VertexQuantileFilter(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "VertexQuantileFilter"





from __future__ import annotations

from keyword import iskeyword
from pyraphtory._codegen._docstring import convert_docstring
from collections import UserString, UserDict
from functools import cached_property


jpype_type_converters = {}
type_converters = {}


def jpype_type_converter(type_name: str):
    def converter(fun):
        jpype_type_converters[type_name] = fun.__name__
        return fun
    return converter


def type_converter(type_name: str):
    def converter(fun):
        type_converters[type_name] = fun.name
        return fun
    return converter


class LazyStr(UserString, str):
    def __new__(cls, *args, **kwargs):
        # we need to extend str so help works but don't actually want to create the immutable str here
        self = super().__new__(cls, "")
        return self

    def __init__(self, seq=None, initial=None):
        self._initial = initial
        if seq is not None:
            super().__init__(seq)

    @cached_property
    def data(self):
        value = self._initial()
        return value


class LazyAnnotations(UserDict, dict):
    # UserDict here is necessary such that dict(LazyAnnotations) actually calls the methods
    # instead of using fast dict copy
    def __init__(self, initials: dict):
        super().__init__(initials)
        self.lazy = set(initials.keys())

    def _unlazyfy(self):
        for key in self.lazy:
            self.__getitem__(key)

    def items(self):
        self._unlazyfy()
        return super().items()

    def values(self):
        self._unlazyfy()
        return super().values()

    def __getitem__(self, item):
        if item in self.lazy:
            self.lazy.remove(item)
            self[item] = clean_type(super().__getitem__(item))
        return super().__getitem__(item)

    def add_lazy_item(self, key, lazy_value):
        self[key] = lazy_value
        self.lazy.add(key)



def clean_identifier(name: str):
    if iskeyword(name):
        return str(name + "_")
    else:
        return str(name)


def clean_type(scala_type):
    from pyraphtory.interop._interop import get_type_repr
    type_name = str(get_type_repr(scala_type))
    return type_name


def get_jpype_type_converter(type):
    name = type.toString()
    return jpype_type_converters.get(name, type_converters.get(name, "to_jvm"))


def get_type_converter(type):
    name = type.toString()
    return type_converters.get(name, "to_jvm")


def build_method(name: str, method, jpype: bool, globals: dict, locals: dict):
    if jpype:
        type_converter = get_jpype_type_converter
    else:
        type_converter = get_type_converter

    params = [clean_identifier(name) for name in method.parameters()]
    types = list(method.types())
    name = clean_identifier(name)
    java_name = clean_identifier(method.name()) if jpype else method.name()
    implicits = [clean_identifier(name) for name in method.implicits()]
    implicit_types = types[len(params):]
    types = types[:len(params)]
    nargs = method.n()
    varargs = method.varargs()
    if varargs:
        varparam = params.pop()
        varparam_type = types.pop()
    defaults = method.defaults()
    required = min(defaults.keys(), default=nargs)

    args = ["self"]
    args.extend(f'{p} = DefaultValue("{defaults[i]}")' if i in defaults
                else (f"{p} = None" if i > required
                      else f"{p}")
                for i, p in enumerate(params))
    if varargs:
        args.append(f"*{varparam}")
    if implicits:
        args.append("_implicits=()")
    args = ", ".join(args)

    lines = [f"def {name}({args}):"]
    if implicits:
        lines.append(f"    if len(_implicits) < {len(implicits)}:")
        lines.append(f"        raise RuntimeError('missing implicit arguments')")
        lines.append(f"    if len(_implicits) > {len(implicits)}:")
        lines.append(f"        raise RuntimeError('too many implicit arguments')")
        for i, p in enumerate(implicits):
            lines.append(f"    {p} = {type_converter(implicit_types[i])}(_implicits[{i}])")
    lines.extend(f"    {p} = _check_default(self._jvm_object, {p}, {type_converter(types[i])})" if i in defaults else
                 f"    {p} = {type_converter(types[i])}({p})" for i, p in enumerate(params))
    if varargs:
        lines.append(f"    {varparam} = make_varargs(to_jvm(list({type_converter(varparam_type)}(v) for v in {varparam})))")
        params.append(varparam)
    lines.append(f"    return to_python(getattr(self._jvm_object, '{java_name}')({', '.join(p for p in params + implicits)}))")
    exec("\n".join(lines), globals, locals)
    py_method = locals[name]
    docstr = method.docs()
    py_method.__doc__ = LazyStr(initial=lambda: convert_docstring(docstr))
    py_method.__annotations__ = LazyAnnotations({n: t for n, t in zip(params, types)})
    if varargs:
        py_method.__annotations__.add_lazy_item(varparam, varparam_type)
    return py_method



from __future__ import annotations

import inspect
import re
import traceback
from abc import ABCMeta, ABC
import os

from typing import *
import cloudpickle as pickle
from functools import cached_property
from threading import Lock, RLock
from copy import copy
from textwrap import indent
from pyraphtory._codegen import _codegen
from pyraphtory._codegen._codegen import jpype_type_converter, LazyStr
from jpype import JObject, JBoolean, JByte, JShort, JInt, JLong, JFloat, JDouble, JString

from pyraphtory._codegen._docstring import convert_docstring


@jpype_type_converter("Long")
def long_converter(value):
    return JLong(value)


@jpype_type_converter("Int")
def int_converter(value):
    return JInt(value)


@jpype_type_converter("Short")
def short_converter(value):
    return JShort(value)


@jpype_type_converter("Byte")
def byte_converter(value):
    return JByte(value)


@jpype_type_converter("Float")
def float_converter(value):
    return JFloat(value)


@jpype_type_converter("Double")
def double_converter(value):
    return JDouble(value)


@jpype_type_converter("Boolean")
def boolean_converter(value):
    return JBoolean(value)


_wrapper_lock = Lock()
_jpype = False


class NoCache:
    def __setitem__(self, key, value):
        pass


_globals = NoCache()


def no_wrapper(jvm_object):
    return jvm_object


_wrappers = {
    "java.lang.Integer": no_wrapper,
    "java.lang.Short": no_wrapper,
    "java.lang.Long": no_wrapper,
    "java.lang.Float": no_wrapper,
    "java.lang.Double": no_wrapper,
}


def repr(obj):
    return str(_scala.repr(obj))


def check_raphtory_logging_env():
    log_level = os.getenv('RAPHTORY_CORE_LOG')
    if log_level is None:
        os.environ["RAPHTORY_CORE_LOG"] = "ERROR"


try:
    from pemja import findClass

    _scala = findClass('com.raphtory.internals.management.PythonInterop')
except (ImportError, ModuleNotFoundError):
    import jpype
    import jpype.config
    jpype.config.onexit = False
    from jpype import JClass
    from pyraphtory import _config

    check_raphtory_logging_env()

    # see https://jpype.readthedocs.io/en/latest/userguide.html#errors-reported-by-python-fault-handler
    try:
        import faulthandler
        faulthandler.enable()
        faulthandler.disable()
    except:
        pass

    jpype.startJVM(_config.java_args, classpath=_config.jars.split(":"))
    _config._jpype_running = True
    from pyraphtory.interop._jpypeinterpreter import JPypeInterpreter, _globals

    _scala = getattr(JClass("com.raphtory.internals.management.PythonInterop$"), "MODULE$")
    _scala.set_interpreter(JPypeInterpreter())
    _jpype = True


def register(cls=None, *, name=None):
    """class decorator for registering wrapper classes.

    Specify a 'name' keyword argument to give the wrapper a name to register it for a range of different types.
    A corresponding match needs to be added on the scala side for the name to have an effect.
    """
    if cls is None:
        return lambda x: register(x, name=name)
    else:
        if name is None:
            name = cls._classname
        if name is None:
            raise ValueError(f"Missing name during registration of {cls!r}")
        _wrappers[name] = cls
        return cls


_JPrimitiveTypes = (JBoolean, JByte, JShort, JInt, JLong, JFloat, JDouble, JString)


def _isJPrimitive(obj):
    return isinstance(obj, _JPrimitiveTypes)


def is_PyJObject(obj):
    """Needed because Pemja objects do not support isinstance"""
    return (type(obj).__name__ == "PyJObject"
            or (isinstance(obj, JObject) and not _isJPrimitive(obj)))


def snake_to_camel(name: str):
    """
    Convert python snake case names to scala-style camelCase preserving leading underscores
    (multiple intermediate underscores are eliminated).
    """
    parts = re.split(r"([^_])", name, 1)
    rest = parts[2].split("_")
    rest[1:] = [v.capitalize() for v in rest[1:]]
    return "".join(parts[:2] + rest)


def camel_to_snake(name: str):
    """Convert scala-style camel-case names to python style snake case"""
    return _scala.camel_to_snake(name)


def decode(obj):
    """call scala decode function to deal with converting java to scala collections"""
    return _scala.decode(obj)


def decode_tuple(obj):
    """call scala decode_tuple function to convert collection to scala tuple"""
    return _scala.decode_tuple(obj)


def get_methods(obj):
    """look up methods for a java object"""
    logger.trace("Finding methods for {obj!r}", obj=obj)
    return _scala.methods(obj)


def get_methods_from_name(name):
    logger.trace("Finding methods for {name} based on name", name=name)
    return _scala.methods_from_name(name)


def get_wrapper_for_name(name: str):
    logger.trace("Retrieving wrapper for {name!r}", name=name)
    wrapper = _wrappers[name]
    logger.trace("Found wrapper for {name!r} based on class name", name=name)
    return wrapper


def get_type_repr(tpe: Any):
    return _scala.get_type_repr(tpe)


def build_wrapper(name: str, obj: Any):
    # Create a new base class for the jvm wrapper and add methods
    with _wrapper_lock:
        if name in _wrappers:
            # a different thread already created the wrapper
            wrapper = _wrappers[name]
            logger.trace("Found wrapper for {name!r} based on class name after initial wait", name=name)
        else:
            # Check if a special wrapper class is registered for the object
            wrap_name = str(_scala.get_wrapper_str(obj))
            if wrap_name in _wrappers:
                # Add special wrapper class to the top of the mro such that method overloads work
                logger.debug("Using wrapper based on name {wrap_name} for {name}", wrap_name=wrap_name, name=name)
                # Note this wrapper is registered automatically as '_classname' is defined
                wrapper = type(name, (_wrappers[wrap_name],), {"_classname": name, "_initialised": False})
            else:
                # No special wrapper registered, can use base wrapper directly
                logger.debug("No wrapper found for name {}, using GenericScalaProxy", name)
                wrapper = type(name, (GenericScalaProxy,), {"_classname": name, "_initialised": False})
            logger.trace("New wrapper created for {name!r}", name=name)
    logger.trace("Wrapper is {wrapper!r}", wrapper=wrapper)
    return wrapper


def get_wrapper(obj):
    """get wrapper class for a java object"""
    name = str(obj.getClass().getName())
    try:
        return get_wrapper_for_name(name)
    except KeyError:
        return build_wrapper(name, obj)


def to_jvm(value):
    """convert wrapped object to underlying jvm representation"""
    if is_PyJObject(value):
        logger.trace("Converting value {value!r}, already PyJObject", value=value)
        return value
    elif isinstance(value, ScalaProxyBase):
        logger.trace("Converting value {value!r}, decoding proxy object", value=value)
        return value.jvm
    elif isinstance(value, Mapping):
        logger.trace("Converting value {value!r}, decoding as Mapping", value=value)
        return decode({to_jvm(k): to_jvm(v) for k, v in value.items()})
    elif callable(value):
        logger.trace("Converting value {value!r}, decoding as Function", value=value)
        return _wrap_python_function(value)
    elif isinstance(value, tuple):
        logger.trace(f"Converting value {value!r}, decoding as Tuple", value=value)
        return decode_tuple([to_jvm(v) for v in value])
    elif (isinstance(value, Iterable)
          and not isinstance(value, str)
          and not isinstance(value, bytes)
          and not isinstance(value, bytearray)):
        logger.trace("Converting value {value!r}, decoding as Iterable", value=value)
        return decode([to_jvm(v) for v in value])
    else:
        logger.trace("Converting value {value!r}, primitive value returned unchanged", value=value)
        return value


def to_python(obj):
    """convert jvm object to python by wrapping it if needed"""
    if is_PyJObject(obj):
        wrapper = get_wrapper(obj)
        logger.trace("Calling wrapper with jvm_object={obj}", obj=obj)
        return wrapper(jvm_object=obj)
    elif isinstance(obj, JString):
        return str(obj)
    else:
        logger.trace("Primitive object {obj!r} passed to python unchanged", obj=obj)
        return obj


def find_class(path: str):
    """get the scala companion object instance for a class path"""
    return _scala.find_class(path)


def assign_id(s: str):
    """call the asign_id function (used by graph builder)"""
    return _scala.assign_id(s)


def make_varargs(param):
    """convert parameter list to varargs-friendly array"""
    return _scala.make_varargs(param)


def _wrap_python_function(fun):
    """take a python function and turn it into a scala function"""
    eval_name = f"wrapped_{id(fun)}"
    wrapped = FunctionWrapper(fun)
    pickle_bytes = pickle.dumps(wrapped)
    _globals[eval_name] = wrapped
    if wrapped.n_args == 1:
        return to_jvm(Function1(pickle_bytes, eval_name))
    elif wrapped.n_args == 2:
        return to_jvm(Function2(pickle_bytes, eval_name))
    else:
        raise ValueError("Only functions with 1 or 2 arguments are currently implemented when passing to scala")


class Logger(object):
    """Wrapper for the java logger"""
    def _format_message(self, msg: str, *args, **kwargs):
        try:
            if args or kwargs:
                msg = msg.format(*args, **kwargs)
        except Exception as e:
            logger.error(f"failed to format {msg=} with {args=} and {kwargs=}, cause: {e}")
        finally:
            return msg

    @cached_property
    def logger(self):
        _logger = _scala.logger()
        level = _logger.level()
        if level < 5:
            self.trace = self.no_op

        if level < 4:
            self.debug = self.no_op

        if level < 3:
            self.info = self.no_op

        if level < 2:
            self.warn = self.no_op

        if level < 1:
            self.error = self.no_op
        return _logger

    def error(self, msg: str, *args, **kwargs):
        self.logger.error(self._format_message(msg, *args, **kwargs))

    def warn(self, msg, *args, **kwargs):
        self.logger.warn(self._format_message(msg, *args, **kwargs))

    def info(self, msg, *args, **kwargs):
        self.logger.info(self._format_message(msg, *args, **kwargs))

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(self._format_message(msg, *args, **kwargs))

    def trace(self, msg, *args, **kwargs):
        self.logger.trace(self._format_message(msg, *args, **kwargs))

    def no_op(self, msg, *args, **kwargs):
        pass


logger = Logger()


class FunctionWrapper(object):
    """class used to interface with python functions from scala"""

    def __init__(self, fun=None):
        self._fun = fun
        self.n_args = len(inspect.getfullargspec(fun).args)

    def eval_from_jvm(self, *args):
        return to_jvm(self(*(to_python(v) for v in args)))

    def __call__(self, *args, **kwargs):
        return self._fun(*args, **kwargs)


class DefaultValue(object):
    """Wrap a scala default value accessor"""

    def __init__(self, method):
        self.method = method

    def __call__(self, obj):
        return getattr(obj, self.method)()

    def __repr__(self):
        return "DefaultValue()"


def _check_default(obj, value, converter):
    if isinstance(value, DefaultValue):
        return value(obj)
    else:
        return converter(value)


class ScalaProxyBase(object):
    """Base class for wrapping jvm objects"""
    _jvm_object = None

    @property
    def jvm(self):
        """The underlying Scala jvm object
        """
        return self._jvm_object

    @classmethod
    def _add_method(cls, name, method_array):
        name = _codegen.clean_identifier(name)
        output = {}
        methods = []
        if len(method_array) > 1:
            for i, method in enumerate(sorted(method_array, key=lambda m: (m.varargs(), m.n()))):
                try:
                    methods.append(_codegen.build_method(f"{name}{i}", method, _jpype, globals(), output))
                except Exception as e:
                    traceback.print_exc()
                    raise e
            methods = list(output.values())
            method = OverloadedMethod(methods, name)
        else:
            method = method_array[0]
            try:
                method = _codegen.build_method(name, method, _jpype, globals(), output)
            except Exception as e:
                traceback.print_exc()
                raise e
        if any(m.implicits() for m in method_array):
            method = WithImplicits(method)
        setattr(cls, name, InstanceOnlyMethod(method))


class JVMBase(ScalaProxyBase):
    _initialised = False
    _classname = None
    _init_lock = RLock()

    @classmethod
    def _init_methods(cls, jvm_object):
        """
        Create method wrappers for all jvm methods (only called the first time a wrapper is created)

        :param jvm_object: java object to wrap
        """
        with cls._init_lock:
            # this should only happen once in the case of multiple threads
            if not cls._initialised:
                logger.debug(f"Initialising uninitialised class {cls.__name__}")
                if jvm_object is None:
                    if cls._classname is not None:
                        methods = get_methods_from_name(cls._classname)
                    else:
                        raise RuntimeError("Cannot initialise methods for class without object or classname")
                else:
                    if cls._classname is None:
                        cls._classname = jvm_object.getClass().getName()
                    logger.trace(f"Getting methods for {jvm_object}")
                    methods = get_methods(jvm_object)
                try:
                    jvm_base_index = next(i for i, b in enumerate(cls.__bases__) if issubclass(b, JVMBase)
                                          and not issubclass(b, GenericScalaProxy))
                except StopIteration:
                    jvm_base_index = len(cls.__bases__)
                base = type(cls.__name__ + "_jvm", (JVMBase,), {"_classname": cls._classname})
                for (name, method_array) in methods.items():
                    base._add_method(name, method_array)
                cls.__bases__ = (*cls.__bases__[:jvm_base_index], base, *cls.__bases__[jvm_base_index + 1:])
                if cls.__doc__ is None:
                    cls.__doc__ = LazyStr(initial=lambda: convert_docstring(_scala.docstring_for_class(cls._classname)))
                cls._initialised = True
            else:
                logger.trace("_init_method called for initialised class {}", cls.__name__)


class GenericScalaProxy(JVMBase):
    """Base class for proxy objects that are not constructable from python

    If a subclass defines a '_classname' attribute, it will be automatically
    registered as the base class for proxy objects of that java class.
    """

    def __repr__(self):
        if self._jvm_object is not None:
            return repr(self._jvm_object)
        else:
            return super().__repr__()

    __str__ = __repr__

    @property
    def classname(self):
        """The name of the underlying jvm class of this wrapper
        """
        if self._classname is None:
            if self._jvm_object is not None:
                self._classname = self._jvm_object.getClass().getName()
                logger.trace(f"Retrieved name {self._classname!r} from java object")
        logger.trace(f"Return name {self._classname!r}")
        return self._classname

    def __call__(self, *args, **kwargs):
        """
        Calling a wrapper calls the `apply` method
        """
        logger.trace(f"{self!r} called with {args=} and {kwargs=}")
        return self.apply(*args, **kwargs)

    def __new__(cls, jvm_object=None):
        """Create a new instance and trigger method initialisation if needed"""
        if jvm_object is not None:
            if not cls._initialised:
                cls._init_methods(jvm_object)
        self = super().__new__(cls)
        self._jvm_object = jvm_object
        return self

    def __init_subclass__(cls, **kwargs):
        """automatically register wrappers that have a '_classname' defined"""
        super().__init_subclass__(**kwargs)
        if cls._classname is not None:
            cls._initialised = False
            register(cls)
            cls._init_methods(None)


class InstanceOnlyMethod(object):
    """Instance method that does not shadow class method of the same name"""

    def __init__(self, method):
        self.__func__ = method
        self.__name__ = method.__name__

    def __set_name__(self, owner, name):
        self.__name__ = name

    def __get__(self, instance, owner=None):
        if instance is None:
            # May shadow class method of the same name!
            try:
                return object.__getattribute__(owner.__class__, self.__name__).__get__(owner, owner.__class__)
            except Exception as e:
                logger.trace("InstanceOnlyMethod non-shadowed due to exception {}", e)
                return self.__func__
        return self.__func__.__get__(instance, owner)

    def __str__(self):
        return f"{self.__name__}"


class OverloadedMethod:
    def __init__(self, methods, name):
        self.__name__ = name
        self._methods = methods
        self.__signature__ = inspect.signature(self.__class__.__call__)
        self.__doc__ = _codegen.LazyStr(
            initial=lambda: f"Overloaded method with alternatives\n\n"
                            + "\n\n".join(
                f"{self.__name__}{str(inspect.signature(m.__get__(m)))}"
                # hack to get signature as if bound method
                + ("\n" + indent(m.__doc__, "    ") if m.__doc__ else "")
                for m in self._methods)
        )

    def __call__(self, *args, **kwargs):
        errors = []
        for method in self._methods:
            try:
                return method(*args, **kwargs)
            except Exception as e:
                logger.trace("call failed for {name} with exception {e}", e=e, name=self.__name__)
                errors.append(e)
        for e in errors:
            print(e)
        raise RuntimeError(f"No overloaded implementations matched for {self.__name__} with {args=} and {kwargs=}")

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        else:
            bound = copy(self)
            bound.__self__ = instance
            bound._methods = [m.__get__(instance, owner) for m in bound._methods]
            return bound


class WithImplicits:
    """Proxy object for scala method with support for default arguments and implicits"""

    def __init__(self, method):
        self.__name__ = method.__name__
        self._method = method
        self._implicits = []
        self.__doc__ = method.__doc__
        self.__wrapped__ = method

    def __call__(self, *args, **kwargs):
        return self._method(*args, **kwargs, _implicits=self._implicits)

    def __getitem__(self, item):
        """support specifying implicit arguments with [val] syntax"""
        if not isinstance(item, tuple):
            item = (item,)
        self._implicits = item
        return self

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            bound = copy(self)
            bound._method = bound._method.__get__(instance, owner)
            return bound


class ScalaObjectProxy(ScalaProxyBase, ABCMeta, type):
    """Metaclass for wrapping Scala companion objects"""
    _base_initialised = False
    _base_init_lock = RLock()

    @property
    def jvm(self):
        """Underlying Scala companion object instance"""
        if self._jvm_object is not None:
            return self._jvm_object
        elif self._classname is not None:
            return find_class(self._classname)
        else:
            return None

    def __subclasscheck__(self, subclass):
        try:
            value = super().__subclasscheck__(subclass)
            return value
        except AttributeError as e:
            return False

    def _from_jvm(cls, jvm_object):
        if cls._base_initialised:
            return cls
        else:
            with cls._base_init_lock:
                cls._init_base_methods(jvm_object)
                cls._jvm_object = jvm_object
            return cls

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Injects an additional specialised type to avoid interference between different classes"""
        concrete = "_classname" in attrs
        if not concrete:
            actual_mcs = mcs
        else:
            actual_mcs = type.__new__(mcs, name + "_", (mcs,), {"_classname": attrs["_classname"]})
            actual_mcs._init_base_methods(_scala.find_class(actual_mcs._classname))

        cls = type.__new__(actual_mcs, name, bases, attrs, **kwargs)
        if concrete:
            cls._jvm_object = _scala.find_class(actual_mcs._classname)
            register(cls._from_jvm, name=cls._classname + "$")

        return cls

    @classmethod
    def _init_base_methods(mcs, jvm_object):
        """add companion object methods to metaclass"""
        with mcs._base_init_lock:
            # this should only happen once in case of multiple threads
            if not mcs._base_initialised:
                logger.trace(f"uninitialised class {mcs.__name__}")
                if jvm_object is None:
                    raise RuntimeError("Need object to find methods")
                logger.trace(f"Getting methods for {jvm_object}")
                methods = get_methods(jvm_object)
                for (name, method_array) in methods.items():
                    mcs._add_method(name, method_array)
                mcs._base_initialised = True


class ScalaClassProxy(GenericScalaProxy, ABC, metaclass=ScalaObjectProxy):
    """Base class for wrapper objects that are constructable from python"""

    @classmethod
    def _build_from_python(cls, *args, **kwargs):
        """Override to control python-side constructor behaviours (e.g., using a builder for sequence construction)"""
        return cls.apply(*args, **kwargs)

    def __new__(cls, *args, jvm_object=None, **kwargs):
        """New instance construction from python uses the `apply` classmethod."""
        if jvm_object is None:
            # call scala constructor
            return cls._build_from_python(*args, **kwargs)
        else:
            # construct from existing object
            self = super().__new__(cls, jvm_object=jvm_object)
            return self


class ScalaClassProxyWithImplicits(ScalaClassProxy):
    _implicits = ()

    @classmethod
    def _build_from_python(cls, *args, **kwargs):
        """Override to control python-side constructor behaviours (e.g., using a builder for sequence construction)"""
        return cls.apply[cls._implicits](*args, **kwargs)

    def __class_getitem__(cls, item):
        new_cls = copy(cls)
        new_cls._implicits = item
        return new_cls


class Function1(ScalaClassProxy):
    """Proxy object for wrapping python functions with 1 argument"""
    _classname = "com.raphtory.internals.management.PythonFunction1"


class Function2(ScalaClassProxy):
    """Proxy object for wrapping python functions with 2 arguments"""
    _classname = "com.raphtory.internals.management.PythonFunction2"


class ScalaPackage(ScalaProxyBase):
    """Proxy object for looking up scala classes based on path
    """

    @property
    def _jvm_object(self):
        return find_class(self._path)

    def __init__(self, path: str):
        self._path = path

    def __call__(self, *args, **kwargs):
        return to_python(self._jvm_object).apply(*args, **kwargs)

    def __getattr__(self, item):
        return ScalaPackage(".".join((self._path, item)))

    def __getitem__(self, item):
        return to_python(self._jvm_object).apply[item]



"""Python wrappers for Sinks (used to output algorithm results to file)"""

from pyraphtory.interop import ScalaClassProxy


class Sink:
    """Base class for Raphtory sinks"""
    pass


class FileSink(ScalaClassProxy, Sink):
    _classname = "com.raphtory.sinks.FileSink"


class PrintSink(ScalaClassProxy, Sink):
    _classname = "com.raphtory.sinks.PrintSink"


# class PulsarSink(ScalaClassProxy, Sink):
#     _classname = "com.raphtory.pulsar.sink.PulsarSink"



"""Multilayer views of the graph"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.temporal.views."


class MultilayerView(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "MultilayerView"



from __future__ import annotations
import platform
from packaging import tags


OS_MAC = 'Darwin'
OS_LINUX = 'Linux'
OS_X64 = 'x64'
OS_AARCH64 = 'aarch64'


def getOS() -> str:
    platform_system = platform.system()
    if platform_system in (OS_MAC, OS_LINUX):
        return platform_system
    else:
        raise Exception("Unsupported OS. Cannot install.")


def getArch() -> str:
    if platform.machine() == "x86_64":
        return OS_X64
    elif platform.machine() == 'arm64':
        return OS_AARCH64
    else:
        raise Exception("Unsupported Architecture. Cannot install.")


def get_platform_tag() -> str:
    if getOS() == OS_MAC:
        if getArch() == OS_X64:
            return str(next(tags.mac_platforms((10, 11))))
        else:
            return str(next(tags.mac_platforms((11, 0))))
    else:
        return str(next(tags.sys_tags()).platform)



"""Algorithms for computing centralities"""

from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy
from pyraphtory.interop import ScalaClassProxyWithImplicits


_prefix = "com.raphtory.algorithms.generic.centrality."


class Assortativity(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Assortativity"


class AverageNeighbourDegree(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "AverageNeighbourDegree"


class Degree(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "Degree"


class Distinctiveness(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "Distinctiveness"


class PageRank(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "PageRank"


class WeightedDegree(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "WeightedDegree"


class WeightedPageRank(ScalaAlgorithm, ScalaClassProxyWithImplicits):
    _classname = _prefix + "WeightedPageRank"


from parsy import string, whitespace, any_char, peek, alt, eof, string_from, Parser, regex, letter, decimal_digit, \
    generate, ParseError

code_conversions = {}


def camel_to_snake(name: str):
    from pyraphtory.interop._interop import camel_to_snake as convert
    from pyraphtory._codegen._codegen import clean_identifier
    if name[0].islower():
        name = convert(name)
    return clean_identifier(name)


def parse_name(name):
    return camel_to_snake(name)


def finalise_param(s):
    return s + ":"


def convert_param(s):
    return ":param"


def convert_returns(s):
    return ":returns:"


def join_tokens(tokens):
    return "".join(tokens)


def ignore(s):
    return ""


def debug(name):
    """useful for printing debug info during parsing"""

    def debug_wrapped(s):
        print(f"{name} parsed {s}")
        return s
    return debug_wrapped


def capitalise(s: str):
    return s.capitalize()


def report_unparsed(s: str):
    print(f"unparsed: {s}")
    return s


def with_optional_whitespace(parser: Parser):
    return whitespace.optional("") + parser + whitespace.optional("")


def linestart_clean(s: str):
    return "\n"


def as_code(s: str):
    return "`" + s + "`"


# whitespace parsers
non_newline_whitespace = regex(r"[^\S\r\n]+")
non_whitespace_char = regex(r"\S")
space = string(" ")
counted_spaces = space.many().map(len)
newline = (string("\n") | string("\r\n") | string("\r")).result("\n")

# scaladoc line start and end comments
start = (non_newline_whitespace.optional("") + string("/**") + space.optional("")).result("")
line_start = start | (non_newline_whitespace.optional("") + string("*") + space.optional("")).result("")
end = (whitespace.optional("") + string("*").at_least(1).concat() + string("/")).result("")
line_end = newline | end



# blank lines
blank_remaining_line = non_newline_whitespace.until(line_end | eof).concat() + line_end.optional("")
blank_line = line_start.optional("") + blank_remaining_line


# block starts
block_start = peek(start).result("") | blank_line.optional("\n")
block_end = (end | eof).result("") | blank_line.optional("\n")

# separate by spaces
token = non_whitespace_char.until(whitespace | line_end | eof, min=1).concat()  # consume non-whitespace characters

# token that should be converted to snake_case
name = token.map(parse_name)

# convert code expressions
code_others = (string("`") | line_end).should_fail(
    "not end of code") >> any_char  # match anything that is not end of code and return unchanged
boolean = string_from('true', 'false').map(capitalise)
string_expr = string("String").result("str")
bool_expr = string("Boolean").result("bool")
int_expr = string_from("Int", "Long", "Integer", "Short").result("int")
float_expr = string_from("Float", "Double").result("float")
dquoted_str = string('"') + ((line_end | string('"')).should_fail("not end of string") >> any_char).many().concat() + string('"')
squoted_str = string("'") + ((line_end | string("'")).should_fail("not end of string") >> any_char).many().concat() + string("'")
method_or_variable_name = (string("_").many().concat()
                           + regex(r"[a-z]")
                           + (letter | decimal_digit).many().concat()
                           + string("_").many().concat()
                           ).map(camel_to_snake)
method_without_brackets = string(".") + method_or_variable_name + peek(whitespace).result("()")
class_name = (string("_").many().concat()
              + regex(r"[A-Z]")
              + (letter | decimal_digit).many().concat()
              + string("_").many().concat()
              )
fun = string("=>").result("->")

code_item = (boolean
             | string_expr
             | int_expr
             | float_expr
             | bool_expr
             | squoted_str
             | dquoted_str
             | method_without_brackets
             | method_or_variable_name
             | class_name
             | fun
             )
code_expr = (code_item | code_others).many().map(join_tokens)
code = string("`") + code_expr + string("`")

code_block_char = line_end.should_fail("not the end of line") >> any_char


def example_tag(indent):
    return line_start >> non_newline_whitespace.at_least(indent) >> string("@example") >> blank_remaining_line.result("")


def example(indent, output_indent):
    @generate("example")
    def example():
        tag = yield example_tag(indent)
        block = yield codeblock(indent, output_indent)
        return ""
    return example


def codeblock_start(indent, output_indent):
    @generate
    def parser():
        l = yield line_start
        s = yield non_newline_whitespace.at_least(indent).result(" "*output_indent)
        start = yield string("{{{").result(".. code-block::\n"+" "*(output_indent+3)+":dedent:\n")
        end = yield blank_remaining_line
        blank_end = yield block_end
        return s + start + end
    return parser


def codeblock_end(indent=0):
    return line_start >> non_newline_whitespace.at_least(indent).concat() + string("}}}").result("\n") + blank_remaining_line


def codeblock_line(output_indent):
    return line_start.result(" "*output_indent) + (code_item | code_block_char).many().concat() + line_end

def codeblock(indent, output_indent):
    @generate("codeblock")
    def codeblock_parser():
        blank = yield block_start
        start = yield codeblock_start(indent, output_indent)
        lines = yield codeblock_line(output_indent+3).until(codeblock_end(indent)).concat()
        end = yield codeblock_end(indent)
        blank_end = yield block_end
        return blank + start + lines + end + blank_end

    return codeblock_parser


# find links (only extracts text for now)
link_string = string("[[") >> any_char.until(string("]]")).concat() << string("]]")

link_value = ((token >> whitespace >> (class_name | method_or_variable_name).map(as_code)) |
              (class_name | method_or_variable_name).map(as_code))


@generate("link")
def link():
    link_str = yield link_string
    try:
        return link_value.parse(link_str)
    except ParseError:
        return link_str


# inline markup
inline = link | code | non_newline_whitespace | token

# lines
remaining_line = inline.until(line_end | eof).concat() + line_end.optional("")
line = line_start.optional("") + remaining_line

# convert tags
def labeled_tag(tag_name, converted_name=None):
    if converted_name is None:
        converted_name = tag_name
    id = string(f"@{tag_name}").result(f":{converted_name}")
    return (id + non_newline_whitespace + name).map(finalise_param)


def field_list_item(tag_name, converted_name=None):
    id_and_label = labeled_tag(tag_name, converted_name)

    @generate(f"{tag_name}_item")
    def list_item():
        leading_spaces = yield line_start >> counted_spaces
        first_line = yield id_and_label + remaining_line
        body = yield indented_block(leading_spaces)
        return " " * leading_spaces + first_line + body

    return list_item

param = field_list_item("param")
throws = field_list_item("throws", "raises")
tparam = field_list_item("tparam").result("")

return_id = string("@return").result(":returns:")

@generate("return")
def return_():
    leading_spaces = yield line_start >> counted_spaces
    first_line = yield return_id + remaining_line
    body = yield indented_block(leading_spaces)
    return " " * leading_spaces + first_line + body


field_list_item = param | tparam | throws | return_

# rst field lists need blank lines before and after
field_list = block_start + field_list_item.at_least(1).concat() + block_end


def indented_line(indent, output_indent):
    """Identify an indented line of a block and return it with converted indent

    :param indent: indent in scaladoc block
    :param output_indent: indent in python docstring block
    """
    return blank_line | (line_start + ((space * indent).result(" " * output_indent) + remaining_line))


def indented_block(leading_spaces, output_indent=None):
    """Parse an indented block of text

    :param leading_spaces: outer indent of the block
    :param output_indent: indent of output (defaults to the indent in the original)
    """
    @generate("indented_block")
    def indented_block_parser():
        result = []
        blanks = yield blank_line.many().concat()
        result.append(blanks)
        first_indent = yield peek(line_start >> counted_spaces).optional(0)
        nonlocal output_indent
        if output_indent is None:
            output_indent = first_indent-leading_spaces
        if first_indent > leading_spaces:
            output = yield (indented_line(first_indent, leading_spaces + output_indent)
                            | indented_blocks(first_indent, output_indent)).many().concat()
            result.append(output)
        return join_tokens(result)
    return indented_block_parser


def directive(name, indent, output_indent, pythonname=None):
    """Parse a directive-style tag

    :param name: name of scala tag is `@name`
    :param pythonname: name in the output (defaults to name)
    """
    if pythonname is None:
        pythonname = name

    @generate(name)
    def directive_parser():
        result = []
        blank = yield block_start
        result.append(blank)
        leading_spaces = yield line_start >> space.at_least(indent).map(len)
        directive_start = yield string("@" + name).result(" " * output_indent + f".. {pythonname}::\n")
        result.append(directive_start)
        rest = yield non_newline_whitespace.many() >> remaining_line  # this picks up any remaining text in the first line
        result.append(" "*(output_indent+3) + rest)
        block = yield indented_block(leading_spaces, output_indent+3)
        result.append(block)
        blank_end = yield block_end
        result.append(blank_end)
        return join_tokens(result)

    return directive_parser


def indented_blocks(indent=0, output_indent=None):
    if output_indent is None:
        output_indent = indent
    return alt(directive("note", indent, output_indent),
               directive("see", indent, output_indent, pythonname="seealso"),
               example(indent, output_indent),
               codeblock(indent, output_indent)
               )


doc_converter = alt(tparam,
                    field_list,
                    indented_blocks(),
                    end,
                    line,
                    ).until(eof).map(join_tokens)


md_code = string("{s}").optional("").result("") + code
md_link_prefix = (string("[](").result("{py:class}`")
                  + string("com.raphtory").result("pyraphtory")
                  + any_char.until(string(")") | line_end).concat()
                  + string(")").result("`")
                  )
md_inline = md_code | md_link_prefix | non_newline_whitespace | token
md_line = line_start.optional("") + md_inline.until(line_end | eof).concat() + line_end.optional("")

md_doc_converter = (end | md_line).until(eof).concat()


def convert_docstring(docs):
    docs = str(docs)
    if docs:
        try:
            if "{s}`" in docs:
                # this is a markdown docstring, use simplified parser
                cleaned = md_doc_converter.parse(docs)
            else:
                cleaned = doc_converter.parse(docs)
            return cleaned.rstrip("\n")
        except ParseError as e:
            print(e)
            return docs
    else:
        return docs



from __future__ import annotations


from setuptools import build_meta as _orig
import sys
from packaging import tags
from pathlib import Path
from ivy import get_and_run_ivy
from java import check_system_dl_java
from sbt_build import make_python_build
from check_platform import get_platform_tag

prepare_metadata_for_build_wheel = _orig.prepare_metadata_for_build_wheel
prepare_metadata_for_build_editable = _orig.prepare_metadata_for_build_editable
get_requires_for_build_wheel = _orig.get_requires_for_build_wheel
get_requires_for_build_sdist = _orig.get_requires_for_build_sdist
get_requires_for_build_editable = _orig.get_requires_for_build_editable


platform = get_platform_tag()


def build_sdist(sdist_directory, config_settings=None):
    make_python_build(rebuild=True)
    return _orig.build_sdist(sdist_directory, config_settings)


def setup_wheel_dependencies(rebuild=False):
    make_python_build(rebuild)
    java_bin = check_system_dl_java()
    get_and_run_ivy(java_bin)


def build_wheel(wheel_directory, config_settings: dict=None, metadata_directory=None):
    print(sys.path)
    setup_wheel_dependencies(rebuild=False)

    # make wheel platform-specific as the jre downloaded will be different
    if config_settings is not None:
        config_settings.setdefault("--global-option", []).extend(["--plat-name", platform])
    else:
        config_settings = {"--global-option": ["--plat-name", platform]}

    name = _orig.build_wheel(wheel_directory, config_settings, metadata_directory)
    return name


def build_editable(wheel_directory, config_settings=None, metadata_directory=None):
    setup_wheel_dependencies(rebuild=True)
    return _orig.build_editable(wheel_directory, config_settings, metadata_directory)



from __future__ import annotations

import os
import tempfile
from download import safe_download_file, Link
from pathlib import Path
import shutil
import logging
import subprocess


from paths import ivy_folder, lib_folder, ivy_bin, build_folder, root_folder

IVY_BIN = Link(
    link='https://github.com/Raphtory/Data/raw/main/apache-ivy-2.5.0-bin.zip',
    checksum='7c6f467e33c28d82f4f8c3c10575bb461498ad8dcabf57770f481bfea59b1e59'
)


def version() -> str:
    with open(root_folder / "version") as f:
        return f.readline()


def download_ivy(download_dir):
    ivy_file_location = safe_download_file(download_dir, IVY_BIN)
    shutil.unpack_archive(ivy_file_location, extract_dir=download_dir)
    return download_dir / "apache-ivy-2.5.0" / "ivy-2.5.0.jar"


def get_and_run_ivy(java: str | Path) -> None:
        ivy_bin.mkdir(exist_ok=True, parents=True)
        ivy_jar = ivy_bin / "ivy-2.5.0.jar"
        if not ivy_jar.is_file():
            with tempfile.TemporaryDirectory() as download_dir:
                download_dir = Path(download_dir)
                shutil.copyfile(download_ivy(download_dir), ivy_jar)

                logging.info(
                    f"IVY downloaded: {download_dir}, input dir: {ivy_folder}, lib dir: {lib_folder}")
        retrieve = str(lib_folder) + "/[organisation].[artifact]-[revision](-[classifier]).[ext]"
        settings = str(build_folder / "ivysettings.xml")
        # retrieve = "."
        with tempfile.TemporaryDirectory() as cache_dir:
            subprocess.check_call(
                [str(java), f"-Divy_dir={ivy_folder}", f"-Divy_tmp_cache={cache_dir}", "-jar", ivy_jar,
                 "-settings", settings, "-dependency", "com.raphtory", "core_2.13", version(),
                 "-retrieve", retrieve, "-sync", "-refresh", "-confs", "runtime"])




"""Python wrappers for input api"""

from pyraphtory.interop import ScalaClassProxy, GenericScalaProxy


class GraphBuilder(ScalaClassProxy):
    _classname = "com.raphtory.api.input.GraphBuilder"


class Type(ScalaClassProxy):
    _classname = "com.raphtory.api.input.Type"


class Property(GenericScalaProxy):
    _classname = "com.raphtory.api.input.Property"


class ImmutableString(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.ImmutableString"


class MutableString(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableString"


class MutableLong(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableLong"

class MutableInteger(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableInteger"


class MutableDouble(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableDouble"


class MutableFloat(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableFloat"


class MutableBoolean(ScalaClassProxy, Property):
    _classname = "com.raphtory.api.input.MutableBoolean"


class Properties(ScalaClassProxy):
    _classname = "com.raphtory.api.input.Properties"



"""
Python wrappers for the raphtory api
"""

# Need to import all wrappers such that they are properly registered
import pyraphtory.api.algorithm
import pyraphtory.api.context
import pyraphtory.api.formats
import pyraphtory.api.graph
import pyraphtory.api.graphstate
import pyraphtory.api.input
import pyraphtory.api.scala
import pyraphtory.api.sinks
import pyraphtory.api.sources
import pyraphtory.api.spouts
import pyraphtory.api.table
import pyraphtory.api.vertex



"""Python wrappers for graph api"""

from __future__ import annotations
from pyraphtory.interop import register, logger, GenericScalaProxy, ScalaClassProxy
from pyraphtory.api.input import Properties, ImmutableString, Type
from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop._interop import ScalaProxyBase


@register(name="Perspective")
class Perspective(GenericScalaProxy):
    _classname = "com.raphtory.api.time.Perspective"


class Alignment(ScalaClassProxy):
    _classname = "com.raphtory.api.analysis.graphview.Alignment"


class PropertyMergeStrategy(ScalaClassProxy):
    _classname = "com.raphtory.api.analysis.visitor.PropertyMergeStrategy"


@register(name="Graph")
class Graph(GenericScalaProxy):
    _classname = "com.raphtory.api.input.Graph"

    def add_vertex(self, update_time: int, src_id: int | str, properties: list[input.Property] = (), vertex_type: str = "", secondary_index: int=None,time_format="yyyy[-MM[-dd[ HH[:mm[:ss[.SSS]]]]]]"):
        """Adds a new vertex to the graph or updates an existing vertex

           :param update_time: timestamp for vertex update
           :param src_id: ID of vertex to add/update
           :param properties: Optionally specify vertex properties for the update (see :py:mod:`~pyraphtory.input` for the
                              available property types)
           :param vertex_type: Optionally specify a type for the vertex
           :param secondary_index: Optionally specify a secondary index that is used to determine the order
                                   of updates with the same `update_time`
           :param time_format: Optionally set the format if providing a datetime String for update_time.
                                   By default this is yyyy-MM-dd HH:mm:ss.SSS
        """
        if secondary_index is None:
            secondary_index = self.index()
        time = update_time
        if isinstance(update_time, str):
            time = self.parse_datetime(time,time_format)

        if isinstance(src_id, str):
            properties += (ImmutableString("name", src_id),)  # comma makes this a tuple and is apparently the fastest way to do things
            super().add_vertex(time, self.assign_id(src_id), Properties(*properties), Type(vertex_type),secondary_index)
        else:
            super().add_vertex(time, src_id, Properties(*properties), Type(vertex_type), secondary_index)

    def add_edge(self, update_time: int, src_id: int | str, dst_id: int | str, properties: list[input.Property] = (),
                 edge_type: str = "", secondary_index: int = None,time_format="yyyy[-MM[-dd[ HH[:mm[:ss[.SSS]]]]]]"):
        """Adds a new edge to the graph or updates an existing edge

        :param update_time: timestamp for edge update
        :param src_id: ID of source vertex of the edge
        :param dst_id: ID of destination vertex of the edge
        :param properties: edge properties for the update (see :py:mod:`~pyraphtory.input` for the available property types)
        :param edge_type: specify a type for the edge
        :param secondary_index: Optionally specify a secondary index that is used to determine the order
                                of updates with the same `update_time`
        :param time_format: Optionally set the format if providing a datetime String for update_time.
                                By default this is yyyy-MM-dd HH:mm:ss.SSS
        """
        if secondary_index is None:
            secondary_index = self.index()
        time = update_time
        if isinstance(update_time, str):
            time = self.parse_datetime(time,time_format)
        source = src_id
        destination = dst_id
        if isinstance(src_id, str):
            source = self.assign_id(src_id)
        if isinstance(dst_id, str):
            destination = self.assign_id(dst_id)
        super().add_edge(time, source, destination, Properties(*properties), Type(edge_type), secondary_index)


@register(name="TemporalGraph")
class TemporalGraph(Graph):
    _classname = "com.raphtory.api.analysis.graphview.TemporalGraph"

    def transform(self, algorithm: algorithm.PyAlgorithm | algorithm.ScalaAlgorithm):
        """Apply an algorithm to the graph

        :param algorithm: Algorithm to apply to the graph
        :return: Transformed graph

        .. note::
           `transform` keeps track of the name of the applied algorithm and clears the message queues at the end of the algorithm
        """
        if isinstance(algorithm, ScalaProxyBase):
            return super().transform(algorithm)
        else:
            return algorithm(self).clear_messages().with_transformed_name(algorithm.name())

    def execute(self, algorithm):
        """Run an algorithm on the graph and return results using the `tabularise` method of the algorithm

        :param algorithm: Algorithm to run
        :return: Table with algorithm results

        .. note::

           `execute` keeps track of the name of the applied algorithm
        """
        if isinstance(algorithm, ScalaProxyBase):
            return super().execute(algorithm)
        else:
            return algorithm.tabularise(self.transform(algorithm))


class DeployedTemporalGraph(TemporalGraph):
    _classname = "com.raphtory.api.analysis.graphview.PyDeployedTemporalGraph"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug("Graph closed using context manager")
        self.close()






"""Algorithms for dynamics on temporal graphs"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.temporal.dynamic."


class GenericTaint(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "GenericTaint"





import sys


def enable_pydev_debug():
    pass


if hasattr(sys, "gettrace") and sys.gettrace() is not None:
    try:
        import pydevd
        pydevd.connected = True
        pydevd.settrace(suspend=False)
        def enable_pydev_debug():
            """
            Use this function to ensure the pydevd debugger (used by e.g. Pycharm)
            works for python code called from a thread created by the jvm.

            """
            pydevd.connected = True
            pydevd.settrace(suspend=False)
    except Exception:
        pass



"""Algorithms for finding motif counts in temporal graphs"""


from pyraphtory.api.algorithm import ScalaAlgorithm
from pyraphtory.interop import ScalaClassProxy


_prefix = "com.raphtory.algorithms.temporal.motif."


class LocalThreeNodeMotifs(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "LocalThreeNodeMotifs"


class MotifAlpha(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "MotifAlpha"


class ThreeNodeMotifs(ScalaAlgorithm, ScalaClassProxy):
    _classname = _prefix + "ThreeNodeMotifs"





import os
from pathlib import Path
from importlib.resources import files
import pyraphtory
import logging
import shutil
import atexit
import sys
from jpype import JClass


extra_jars = []
_jpype_running = False


class ExitHooks(object):
    """Control python exit while preserving exit code.

    see: https://stackoverflow.com/a/73277719
    """
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        """
        hook into system exit to trap exit code and exceptions
        """
        self._orig_exit = sys.exit
        self._orig_exc_handler = sys.excepthook
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        """
        exit hook that traps the exit code and stores it
        :param code: exit code
        """
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        """
        excepthook that traps the exception and stores it
        """
        self.exception = exc
        self._orig_exc_handler(exc_type, exc, *args)

def exit_handler():
    """Forwards exit code to java and calls java exit if jvm was started"""
    try:
        if _jpype_running:
            System = JClass("java.lang.System")
            if hooks.exit_code is not None:
                System.exit(hooks.exit_code)
            elif hooks.exception is not None:
                System.exit(1)
            else:
                System.exit(0)
    except Exception as e:
        # Catch any error, otherwise we have an infinite recursion
        print(f"Java exit handler raised exception {e!r} which cannot be thrown")


hooks = ExitHooks()
hooks.hook()
atexit.register(exit_handler)


def get_java_home() -> Path:
    logging.info("Getting JAVA_HOME")
    home = os.getenv('JAVA_HOME')
    if home is not None:
        return Path(home)
    elif shutil.which('java') is not None:
        logging.info(f'JAVA_HOME not found. But java found. Detecting home...')
        # resolve JAVA_HOME in case it is a symlink
        home = Path(shutil.which('java')).resolve().parents[1]
        os.environ["JAVA_HOME"] = str(home)
        return home
    else:
        raise FileNotFoundError("JAVA_HOME has not been set, java was also not found")


def get_local_jre_loc() -> Path:
    if os.environ.get("PYRAPHTORY_USE_SYSTEM_JAVA", ""):
        return get_java_home()
    else:
        jre = files(pyraphtory) / "jre"
        if not isinstance(jre, Path):
            raise RuntimeError("Pyraphtory is not installed correctly, are you trying to import from a compressed file?")

        os.environ["JAVA_HOME"] = str(jre)
        return jre


def get_local_lib() -> Path:
    lib = files(pyraphtory) / "lib"
    if not isinstance(lib, Path):
        raise RuntimeError("Pyraphtory is not installed correctly, are you trying to import from a compressed file?")
    return lib


def get_local_jar_path():
    lib = get_local_lib()
    return str(lib) + "/*"


def join_jar_path(path: str, *new_paths: str) -> str:
    return ":".join((path, ":".join(new_paths)))


def setup_raphtory_jars():
    custom_jar_path = os.environ.get("PYRAPHTORYPATH", "")
    java_args_env = os.environ.get("PYRAPTHORY_JVM_ARGS", "")
    path = get_local_jar_path()
    if custom_jar_path:
        path = join_jar_path(path, custom_jar_path)
    return path, java_args_env


jre = get_local_jre_loc()
jars, java_args = setup_raphtory_jars()
java = str(jre / "bin" / "java")


def add_classpath(path: str):
    """
    Add custom jars or java classes.

    Note that this function only has an effect before importing the `algorithms`,
    `api`, or `interop` modules or calling the `local` or `remote` function to start a Raphtory context.

    :param path: Path to add to the jvm classpath. This supports the wildcard '*' to add all files in a folder. Use ':'
                 to separate multiple files/folders.
    """
    global jars
    jars = join_jar_path(jars, path)


def get_java_args():
    """
    Return the arguments passed to the jvm at startup
    """
    return java_args


def set_java_args(args: str):
    """
    Set the arguments passed to the jvm at startup. 

    :param args: New arguments to pass

    Note this function overrides any existing settings passed in via the environment. 
    These arguments have no effect after the jvm is started (i.e., after importing the `algorithms`,
    `api`, or `interop` modules or calling the `local` or `remote` function to start a Raphtory context).
    """
    global java_args
    java_args = args



"""Classes and methods for interacting with the jvm

.. data:: logger

   Wrapper for the java logger
"""
from ._interop import ScalaClassProxy, GenericScalaProxy, ScalaPackage, ScalaClassProxyWithImplicits, logger, register, to_jvm, to_python

__all__ = [
    "ScalaClassProxy",
    "GenericScalaProxy",
    "ScalaPackage",
    "logger",
    "register"
]
